import contextlib
import copy
import dataclasses
import datetime
import functools
import inspect
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
import time
import warnings

import numpy as np


# Number of times to repeat each performance measurement
repeats = 5

# Rough number of seconds to spend on each performance measurement
t_perf = 0.2

# JITs known by this module
names = {
    'python': 'Python',
    'numpy': 'NumPy',
    # JITs
    'cycept': 'Cycept',
    'cython': 'Cython',
    'numba': 'Numba',
}
class Jit:
    def __init__(self, name, version, *decorators):
        self.name = name
        self.version = version
        self.decorators = decorators
@dataclasses.dataclass
class Findings:
    python: float = np.inf
    numpy: float = np.inf
    cycept: float = np.inf
    cython: float = np.inf
    numba: float = np.inf


# Function returning the JITs available on the system
@functools.cache
def get_jits():
    include = lambda name: not setup['jits'] or name in setup['jits']
    jits = {}
    # Cycept
    if include('cycept'):
        try:
            import cycept
        except Exception:
            pass
        else:
            jits['cycept'] = Jit(
                'cycept',
                cycept.__version__,
                functools.partial(cycept.jit, silent=setup['silent']),
            )
    # Cython
    if include('cython'):
        try:
            import cython
        except Exception:
            pass
        else:
            jits['cython'] = Jit('cython', cython.__version__, cython.compile)
            # Set CYTHON_CACHE_DIR to a new temporary directory,
            # ensuring that Cython does not reuse previously compiled modules.
            var = 'CYTHON_CACHE_DIR'
            if var not in os.environ:
                os.environ[var] = tempfile.mkdtemp()
    # Numba (nopython mode, then fallback to object mode)
    if include('numba'):
        try:
            import numba
        except Exception:
            pass
        else:
            jits['numba'] = Jit(
                'numba',
                numba.__version__,
                functools.partial(
                    numba.jit,
                    nopython=True,
                    fastmath=True,
                ),
                functools.partial(
                    numba.jit,
                    forceobj=True,
                ),
            )
    # Print out information
    print('Python', '.'.join(map(str, sys.version_info[:3])))
    if setup['numpy']:
        print('NumPy', np.__version__)
    print('JITs:')
    for jit in jits.values():
        print('*', names[jit.name], jit.version)
    return jits


# Function for running all performance tests
def bench(show_func=False, silent=True):

    @contextlib.contextmanager
    def set_globals():
        setup_backup = copy.deepcopy(setup)
        # Include every JIT found on the system
        setup['jits'].clear()
        # Include NumPy
        setup['numpy'] = True
        # Disable test asserts
        setup['asserts'] = False
        # If show_func, enable printing of the tested source code
        setup['show_func'] = show_func
        # If silent, disable messages during jitting
        setup['silent'] = silent
        try:
            yield
        except Exception:
            raise
        finally:
            setup.clear()
            setup.update(setup_backup)

    # Discover and run test functions within this module
    with set_globals():
        for var, val in globals().copy().items():
            if not var.startswith('test_') and not var.endswith('_test'):
                continue
            if callable(val):
                val()
        # Cleanup
        jits = get_jits()
        if 'cython' in jits:
            cache_dir = os.environ.get('CYTHON_CACHE_DIR')
            if cache_dir is not None and pathlib.Path(cache_dir).is_dir():
                shutil.rmtree(cache_dir, ignore_errors=True)

# Default setup, temporarily changed by the bench() function
setup = {
    'jits': ['cycept'],
    'numpy': False,
    'asserts': True,
    'show_func': False,
    'silent': True,
}


# Main function for performing the performance measurements
def perf(func, *args, **kwargs):
    func_numpy = None
    if isinstance(func, tuple):
        func, func_numpy = func
    prepare = None
    for i, arg in enumerate(args):
        if callable(arg) and getattr(arg, '__name__') == 'prepare':
            prepare = arg
            args = args[:i] + args[i + 1:]
            break
    else:
        for key, val in kwargs.items():
            if key == 'prepare' and callable(val):
                prepare = val
                kwargs.pop(key)
                break

    def measure(name, func):
        calls = determine_calls(func)
        t_best = np.inf
        for _ in range(repeats):
            result, t = run(func, calls)
            t_best = min(t_best, t)
        setattr(results, name, result)
        setattr(timings, name, t_best/calls)
        return calls

    def determine_calls(func):
        result, t = run(func)
        calls = max(int(float(f'{t_perf/t:.0e}')), 1)
        return calls

    def run(func, calls=1):
        if prepare is None:
            tic = time.perf_counter()
            for _ in range(calls):
                result = func(*args, **kwargs)
            toc = time.perf_counter()
            return result, toc - tic
        t_prepare = 0
        tic = time.perf_counter()
        for _ in range(calls):
            tic_prepare = time.perf_counter()
            args_prepare = prepare()
            toc_prepare = time.perf_counter()
            t_prepare += toc_prepare - tic_prepare
            result = func(*args, *args_prepare, **kwargs)
        toc = time.perf_counter()
        return result, (toc - tic) - t_prepare

    def print_timings(name, calls):
        def check_equal():
            result = getattr(results, name)
            result_python = results.python
            same_types = (
                isinstance(result, (int, np.integer))
                and isinstance(result_python, (int, np.integer))
            ) or (
                isinstance(result, (float, np.floating))
                and isinstance(result_python, (float, np.floating))
            ) or type(result) is type(result_python)
            is_equal = False
            if same_types:
                if isinstance(result, np.ndarray):
                    if issubclass(result.dtype.type, np.floating):
                        is_equal = np.allclose(result, result_python)
                    else:
                        is_equal = (result == result_python).all()
                else:
                    if isinstance(result, (float, np.floating)):
                        is_equal = np.isclose(result, result_python)
                    else:
                        is_equal = (result == result_python)
            return is_equal
        def get_speedup():
            t = getattr(timings, name)
            speedup = timings.python/t
            if speedup >= 10:
                return f'{speedup:.0f}'
            elif speedup >= 1:
                return f'{speedup:.1f}'
            return f'{speedup:.2f}'
        maxlen = max(map(len, names.values()))
        s0 = f'⚡ {{:<{maxlen + 1}}}'.format(names[name] + ':')
        t = getattr(timings, name)
        if t == np.inf:
            print(f'{s0} Fails to compile'.replace('⚡', '❌'))
            return
        if name != 'python':
            is_equal = check_equal()
            if not is_equal:
                print(f'{s0} Disagrees with pure Python'.replace('⚡', '❓'))
            if setup['asserts'] and name == 'cycept':
                assert is_equal
            if not is_equal:
                return
        s1 = 's' if calls > 1 else ' '
        s2 = ''
        if name != 'python':
            speedup = get_speedup()
            s2 = f' ({speedup}x)'
        calls_pretty = pretty_int(calls)
        t_pretty = pretty_time(t)
        print(
            f'{s0} {calls_pretty} loop{s1}, best of {repeats}: '
            f'{t_pretty} per loop{s2}'
        )

    def print_heading():
        def dedent(text, indentation='auto'):
            lines = text.split('\n')
            if indentation == 'auto':
                indentation = ' ' * (len(lines[0]) - len(lines[0].lstrip()))
            lines = [
                l for line in lines
                if (l := line.removeprefix(indentation).rstrip())
            ]
            text = '\n'.join(lines)
            return text
        def get_source(func):
            source = dedent(inspect.getsource(func))
            return source
        caller_name = inspect.currentframe().f_back.f_back.f_code.co_name
        hashes = '#' * (1 + setup['show_func'])
        print(f'\n{hashes} {caller_name}')
        if not setup['show_func']:
            return
        doc = dedent(globals()[caller_name].__doc__, ' ' * 4)
        print('\n'.join(f'# {line}' for line in doc.split('\n')))
        source = get_source(func)
        print(source)
        if func_numpy is not None:
            source_numpy = get_source(func_numpy)
            print(source_numpy)
    jits = get_jits()
    print_heading()
    timings = Findings()
    results = Findings()
    # Pure Python
    calls = measure('python', func)
    print_timings('python', calls)
    # NumPy
    if setup['numpy'] and func_numpy is not None:
        calls = measure('numpy', func_numpy)
        print_timings('numpy', calls)
    # JITs
    for name, jit in jits.items():
        for jit_decorator in jit.decorators:
            func_jitted = jit_decorator(func)
            try:
                with silence():
                    run(func_jitted)
            except Exception:
                if name == 'cycept':
                    raise
            else:
                calls = measure(name, func_jitted)
                break
        print_timings(name, calls)
    return timings


# Context manager for silencing stdout and stderr, Python and compiler warnings
@contextlib.contextmanager
def silence():

    @contextlib.contextmanager
    def hack_subprocess_Popen():
        subprocess_Popen = subprocess.Popen
        def Popen(*args, **kwargs):
            kwargs['stdout'] = subprocess.DEVNULL
            kwargs['stderr'] = subprocess.DEVNULL
            return subprocess_Popen(*args, **kwargs)
        subprocess.Popen = Popen
        try:
            yield
        except Exception:
            raise
        finally:
            subprocess.Popen = subprocess_Popen

    if not setup['silent']:
        yield
        return
    with (
        (devnull := open(os.devnull, 'w')),
        contextlib.redirect_stdout(devnull),
        contextlib.redirect_stderr(devnull),
        warnings.catch_warnings(),
        hack_subprocess_Popen(),
    ):
        warnings.simplefilter('ignore')
        yield

# Function for converting int of one significant digit to pretty str
def pretty_int(n):
    superscripts = '⁰¹²³⁴⁵⁶⁷⁸⁹'
    if n < 10_000:
        pretty_i =  str(n)
    else:
        factor, _, exponent = f'{n:.0e}'.partition('e')
        pretty_i = f'{factor}×10{superscripts[int(exponent)]}'
    return f'{pretty_i:>5}'

# Function for converting time interval in seconds to pretty str
def pretty_time(t):
    if t <= 0:
        return 'no time at all'
    if t == np.inf:
        return '∞'
    units = {
        'ns': 1e-9,
        'μs': 1e-6,
        'ms': 1e-3,
        's': 1,
    }
    for unit, ratio in units.items():
        factor = 59.95 if unit == 's' else 999.5
        if t < factor*ratio:
            num = f'{t/ratio:#.3g}'.rstrip('.')
            t_pretty = f'{num} {unit}'
            break
    else:
        t_pretty = str(datetime.timedelta(seconds=int(round(t)))).removeprefix('0:')
    return f'{t_pretty:>7}'


# Test functions below

def test_prime():
    """Compute the n'th prime number.
    This tests the performance of integer operations.
    """
    def f(n):
        count = 0
        i = 1
        while True:
            i += 1
            for j in range(2, i):
                if i % j == 0:
                    break
            else:
                count += 1
                if count == n:
                    return i
    n = 500
    timings = perf(f, n)
    if setup['asserts']:
        assert timings.cycept < timings.python / 4


def test_wallis():
    """Compute π using the Wallis product.
    This tests the performance of integer and floating-point operations.
    """
    def f(n):
        i : np.int64
        π = 2
        for i in range(1, n):
            π *= 4 * i ** 2 / (4 * i ** 2 - 1)
        return π
    def f_numpy(n):
        a = 4 * np.arange(1, n, dtype=np.int64) ** 2
        return 2 * np.prod(a / (a - 1))
    n = 2_000_000
    timings = perf((f, f_numpy), n)
    if setup['asserts']:
        assert timings.cycept < timings.python / 50


def test_leibniz():
    """Compute π using the Leibniz formula.
    This tests the performance of floating-point operations.
    """
    def f(n):
        π = 1
        sgn = 1
        for i in range(3, n, 2):
            sgn = -sgn
            π += sgn / i
        π *= 4
        return π
    def f_numpy(n):
        denom = np.arange(1, n, 2, dtype=np.int64)
        num = np.ones_like(denom)
        num[1::2] = -1
        return 4 * (num / denom).sum()
    n = 3_000_000
    timings = perf((f, f_numpy), n)
    if setup['asserts']:
        assert timings.cycept < timings.python / 10


def test_fibonacci():
    """Compute the n'th Fibonacci number using recursion.
    This tests the performance of recursive function calls.
    """
    def f(n):
        if n < 2:
            return n
        return f(n - 2) + f(n - 1)
    n = 30
    timings = perf(f, n)
    if setup['asserts']:
        assert timings.cycept < timings.python / 25


def test_mostcommon():
    """Find the most common object in a list.
    This tests the performance of pure Python operations.
    """
    def f(objs) -> object:
        obj: object
        objs = objs.copy()
        counter = {}
        while objs:
            obj = objs.pop()
            if not obj:
                continue
            if obj not in counter:
                counter[obj] = 0
            counter[obj] += 1
        n = max(counter.values())
        for obj, count in counter.items():
            if count == n:
                return obj
    n = 100
    objs =  1 * n * list(np.arange(n, dtype=np.int64))
    objs += 2 * n * list(np.linspace(0, 1, n, dtype=np.float64))
    objs += 3 * n * [None, 'hello', True, (0, 1, 2), 'hello']
    timings = perf(f, objs)
    if setup['asserts']:
        assert timings.cycept < timings.python


def test_life():
    """Evolve a glider in Conway's Game of Life.
    This tests the performance of pure Python operations and closures.
    """
    def f(n):
        glider = {(0, 2), (1, 0), (1, 2), (2, 1), (2, 2)}
        def evolve(state):
            get_neighbors = lambda x, y: {
                (x + dx, y + dy)
                for dx in range(-1, 2)
                for dy in range(-1, 2)
                if not dx == dy == 0
            }
            squares = state.copy()
            for x, y in state:
                squares |= get_neighbors(x, y)
            state_new = set()
            for x, y in squares:
                neighbors = get_neighbors(x, y)
                count = len(neighbors & state)
                if count == 3 or (count == 2 and (x, y) in state):
                    state_new.add((x, y))
            return state_new
        state = glider.copy()
        for _ in range(4 * n):
            state = evolve(state)
        return state
    n = 30
    timings = perf(f, n)
    if setup['asserts']:
        assert timings.cycept < timings.python


def test_array_0():
    """Compute the value sum((a - b)**2) with a and b being arrays.
    This tests the performance of array indexing.
    """
    def f(a, b):
        x = 0
        for i in range(a.shape[0]):
            for j in range(a.shape[1]):
                x += (a[i, j] - b[i, j]) ** 2
        return x
    def f_numpy(a, b):
        return ((a - b)**2).sum()
    m, n = 1_300, 1_400
    a = np.linspace(0, 1, m * n, dtype=np.float64).reshape((m, n))
    b = np.linspace(1, 0, m * n, dtype=np.float64).reshape((m, n))
    timings = perf((f, f_numpy), a, b)
    if setup['asserts']:
        assert timings.cycept < timings.python / 50


def test_array_1():
    """Transform array in-place as a[i, j] /= i**2 + j**2.
    This tests the performance of array indexing.
    """
    def f(a):
        for i in range(a.shape[0]):
            for j in range(i == 0, a.shape[1]):
                a[i, j] /= i**2 + j**2
        return a
    def f_numpy(a):
        i2 = np.arange(a.shape[0]) ** 2
        j2 = np.arange(a.shape[1]) ** 2
        k2 = i2[:, None] + j2[None, :]
        k2[0, 0] = 1
        a /= k2
        return a
    def prepare():
        m, n = 1_100, 1_300
        a = np.linspace(0, 1, m * n, dtype=np.float64).reshape((m, n))
        return a,
    timings = perf((f, f_numpy), prepare)
    if setup['asserts']:
        assert timings.cycept < timings.python / 50


def test_matmul():
    """Compute the matrix multiplication a @ b.
    This tests the performance of array indexing.
    Here NumPy is expected to be the fastest due to its
    much more sophisticated implementation.
    """
    def f(a, b):
        m, n = a.shape
        p, q = b.shape
        b = b.transpose().copy()
        c = np.empty((m, q), dtype=a.dtype)
        for i in range(m):
            for j in range(q):
                val = 0
                for k in range(n):
                    val += a[i, k] * b[j, k]
                c[i, j] = val
        return c
    def f_numpy(a, b):
        return a @ b
    m, n = 150, 350
    p, q = n, 250
    a = np.linspace(0, 1, m * n, dtype=np.float64).reshape((m, n))
    b = np.linspace(0, 1, p * q, dtype=np.float64).reshape((p, q))
    timings = perf((f, f_numpy), a, b)
    if setup['asserts']:
        assert timings.cycept < timings.python / 200


def test_mandelbrot():
    """Compute an image of the Mandelbrot set.
    This tests the performance of complex numbers and iteration.
    """
    def f(x_min, x_max, y_min, y_max, n_max, image):
        for i in range(image.shape[0]):
            for j in range(image.shape[1]):
                x = x_min + i*(x_max - x_min)/(image.shape[0] - 1)
                y = y_min + j*(y_max - y_min)/(image.shape[1] - 1)
                c = x + y*1j
                z = 0j
                for n in range(n_max):
                    z = z * z + c
                    if abs(z) >= 2:
                        break
                image[i, j] = n
        return image
    x_min, x_max = -2, 0.5
    y_min, y_max = -1.2, 1.2
    n_max = 30
    width = 300
    height = int(width*(y_max - y_min)/(x_max - x_min))
    image = np.empty((width, height), dtype=np.uint8)
    timings = perf(f, x_min, x_max, y_min, y_max, n_max, image)
    if setup['asserts']:
        assert timings.cycept < timings.python / 30


def test_trapz():
    """Compute a definite integral using the trapezoidal rule.
    This tests the performance of floating-point operations.
    """
    def f(x, y):
        s = 0
        for i in range(1, x.shape[0]):
            s += (y[i - 1] + y[i]) * (x[i] - x[i - 1])
        s /= 2
        return s
    def f_numpy(x, y):
        return np.trapz(y, x)
    n, m = 100_001, 20
    x = np.linspace(0, n * np.pi, n * m)
    x += np.sin(x)
    y = np.sin(x)
    timings = perf((f, f_numpy), x, y)
    if setup['asserts']:
        assert timings.cycept < timings.python / 200


def test_nbody():
    """Perform an N-body simulation.
    This tests the performance of array indexing
    and floating-point operations.
    """
    def f(r, v, eps, steps, dt=1e-2):
        n, ndim = r.shape
        d = np.empty(3, dtype=np.float64)
        for t in range(steps):
            for i in range(n):
                for j in range(i + 1, n):
                    r2 = 0
                    for dim in range(ndim):
                        d[dim] = r[i, dim] - r[j, dim]
                        r2 += d[dim] ** 2
                    r3_inv_softened = (r2 + eps ** 2) ** (-1.5)
                    for dim in range(ndim):
                        f = -d[dim] * r3_inv_softened
                        v[i, dim] += f
                        v[j, dim] -= f
            for i in range(n):
                for dim in range(ndim):
                    r[i, dim] += v[i, dim] * dt
    def prepare():
        n = 8
        ndim = 3
        eps = 0.1
        steps = 2
        r = (
            np.asarray(np.meshgrid(*[np.arange(n, dtype=np.float64)]*ndim))
            .reshape((ndim, n**ndim)).T.copy()
        )
        r += eps*np.sin(2*np.pi*r)
        v = np.zeros_like(r)
        return r, v, eps, steps
    timings = perf(f, prepare)
    if setup['asserts']:
        assert timings.cycept < timings.python / 200




import contextlib
import logging
import os
import pathlib
import sys
import warnings

import Cython.Build
import Cython.Distutils
import setuptools


# Method for handling Cythonization and C compilation
def compile(module_path, optimizations, c_lang, html):
    module_path = pathlib.Path(module_path)
    optimization_options = OptimizationOptions(optimizations)
    macros = get_macros()
    # Context manager to redirect all logging to stdout
    @contextlib.contextmanager
    def print_logging():
        root = logging.getLogger()
        level = root.level
        root.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        root.addHandler(handler)
        try:
            yield
        except Exception:
            raise
        finally:
            root.removeHandler(handler)
    # Cythonize and compile extension module
    with (
        contextlib.chdir(module_path.parent),
        print_logging(),
    ):
        print(f'Changed working directory to {os.getcwd()}')
        extension_kwargs = {}
        if c_lang:
            extension_kwargs |= {'language': c_lang}
        extension = Cython.Distutils.Extension(
            name=module_path.stem,
            sources=[module_path.with_suffix('.pyx').name],
            define_macros=[(macro, None) for macro in macros],
            extra_compile_args=optimization_options.all,
            extra_link_args=optimization_options.base,
            **extension_kwargs,
        )
        build_ext = get_build_ext()
        build_ext.build_temp = ''
        build_ext.build_lib  = ''
        build_ext.extensions = Cython.Build.cythonize(
            [extension],
            compiler_directives={'language_level': 3},
            annotate=html,
        )
        build_ext.run()


# Function for obtaining a newly instantiated build_ext
def get_build_ext():
    dist = setuptools.dist.Distribution()
    dist.parse_config_files(dist.find_config_files())
    build_ext = Cython.Distutils.build_ext(dist)
    build_ext.finalize_options()
    return build_ext


# Function for detecting the type of C compiler to be used by setuptools
def get_compiler():
    # We adopt compiler names from the compiler_type attributes
    # on the distutils compiler classes. In particular:
    # * 'unix': GCC or Clang
    # * 'msvc': Microsoft Visual C++
    compiler_default = 'unix'
    build_ext = get_build_ext()
    build_ext.extensions = [None]
    try:
        build_ext.run()
    except setuptools.distutils.errors.DistutilsError:
        pass
    if build_ext.compiler is None:
        return compiler_default
    return build_ext.compiler.compiler_type


# Class for handling compiler optimization options
class OptimizationOptions:

    defaults = {
        'base': True,
        'fastmath': True,
        'native': True,
    }
    compilers = {'unix', 'msvc'}
    mappings = {
        'base': {
            'unix': ['-O3', '-funroll-loops'],
            'msvc': ['/O2'],
        },
        'fastmath': {
            'unix': ['-ffast-math'],
            'msvc': ['/fp:fast'],
        },
        'native': {
            'unix': ['-march=native', '-mtune=native'],
            'msvc': [],
        },
    }

    def __init__(self, optimizations=None):
        if optimizations is None:
            optimizations = {}
        # Merge user and default optimizations
        self.optimizations = self.defaults | optimizations
        if unrecognized := set(optimizations) - set(self.defaults):
            raise AttributeError(
                'Unrecognized optimization(s): {}'.format(', '.join(unrecognized))
            )
        # Set optimization options
        self.compiler = get_compiler()
        if self.compiler not in self.compilers and any(self.optimizations.values()):
            warnings.warn(
                (
                    f'Will apply compiler optimizations for unknown '
                    f'compiler \'{self.compiler}\' in Unix style'
                ),
                RuntimeWarning,
            )
            self.compiler = 'unix'
        for optimization, enable in self.optimizations.items():
            attr = []
            if enable:
                attr = self.mappings[optimization][self.compiler]
            setattr(self, optimization, attr)

    @property
    def all(self):
        return self.base + self.fastmath + self.native


# Function for setting up macros
def get_macros(macros=None):
    if macros is None:
        macros = []
    return macros_default + macros
macros_default = ['CYTHON_WITHOUT_ASSERTIONS', 'NDEBUG']




import ast
import collections
import contextlib
import inspect
import itertools
import pathlib
import re
import subprocess
import sys
import tempfile
import time
import typing
import warnings
import webbrowser


# Class for storing and dynamically handling
# various aspects of the function call to be jitted.
class FunctionCall:

    class Compiled(typing.NamedTuple):
        func: object
        module: object
        source: str
        source_ext: str
        html: str

    class Time(typing.NamedTuple):
        compile: float
        total: float

    def __init__(self, func, wrapper, args, kwargs):
        # Function and call arguments
        self.func = func
        self.wrapper = wrapper
        self.args = args
        self.kwargs = kwargs
        # Properties to be lazily constructed
        self._func_name = None
        self._signature = None
        self._arguments = None
        self._arguments_types = None
        self._hash = None
        self._source = None
        self._ast = None
        self._annotations = None
        self._types = None
        self._module = None
        self._module_dict = None
        self._globals = None
        self._nonlocals = None
        # Record of inferred types of local variables
        self.locals_types = {}
        # Local names which should be disregarded when
        # creating typed local variables.
        self.locals_excludes = {'return'}
        # Additional source lines to be included in the Cython extension module
        self.cython_module_lines = []
        # Compilation products
        self.compiled = None
        # Time measurements
        self.time = None

    # Name of function (with lambda functions renamed)
    @property
    def func_name(self):
        if self._func_name is not None:
            return self._func_name
        self._func_name = self.func.__name__
        if self._func_name == '<lambda>':
            self._func_name = f'_cycept_lambda_{self.hash}'
        return self._func_name

    # Function signature (no type information)
    @property
    def signature(self):
        if self._signature is not None:
            return self._signature
        self._signature = inspect.signature(self.func)
        return self._signature

    # Argument values (dict)
    @property
    def arguments(self):
        if self._arguments is not None:
            return self._arguments
        bound = self.signature.bind(*self.args, **self.kwargs)
        bound.apply_defaults()
        self._arguments = bound.arguments
        return self._arguments

    # Argument types (tuple)
    @property
    def arguments_types(self):
        if self._arguments_types is not None:
            return self._arguments_types
        from .core import get_type
        self._arguments_types = tuple(
            get_type(val) for val in self.arguments.values()
        )
        return self._arguments_types

    # Hash of the function call
    @property
    def hash(self):
        if self._hash is not None:
            return self._hash
        self._hash = abs(hash((self.func, self.wrapper, self.arguments_types)))
        return self._hash

    # Source code of the function
    @property
    def source(self):
        if self._source is not None:
            return self._source
        self._source = self.get_source()
        # Separate annotations from source
        self._annotations, self._source = self.extract_annotations()
        # Invalidate AST
        self._ast = None
        return self._source

    @property
    def ast(self):
        if self._ast is not None:
            return self._ast
        self._ast = ast.parse(self.source)
        return self._ast

    # User-defined annotations within the function
    @property
    def annotations(self):
        if self._annotations is not None:
            return self._annotations
        self._annotations, _ = self.extract_annotations()
        return self._annotations

    # Types to be used during compilation
    @property
    def types(self):
        if self._types is not None:
            return self._types
        from .core import convert_to_cython_type
        # Transform recorded types of locals to str
        # representations of corresponding Cython types.
        locals_types = {
            name: convert_to_cython_type(tp)
            for name, tp in self.locals_types.items()
        }
        # Transform annotations to str representations
        # of Cython types.  Keep unrecognized types as is.
        annotations = {
            name: convert_to_cython_type(tp, tp)
            for name, tp in self.annotations.items()
        }
        # Merge recorded types and annotations.
        # On conflict we keep the annotation.
        self._types = locals_types | annotations
        return self._types

    # Module within which function was defined
    @property
    def module(self):
        if self._module is not None:
            return self._module
        self._module = inspect.getmodule(self.func)
        return self._module

    # Module dictionary associated with function module
    @property
    def module_dict(self):
        if self._module_dict is not None:
            return self._module_dict
        self._module_dict = getattr(self.module, '__dict__', None)
        if self._module_dict is None:
            raise ModuleNotFoundError(
                f'Failed to get __dict__ of module {self.module} '
                f'within which function {self.func} is defined'
            )
        return self._module_dict

    # Global names referenced within the function
    @property
    def globals(self):
        if self._globals is not None:
            return self._globals
        self._globals = inspect.getclosurevars(self.func).globals
        return self._globals

    # Non-local names referenced within the function
    @property
    def nonlocals(self):
        if self._nonlocals is not None:
            return self._nonlocals
        self._nonlocals = inspect.getclosurevars(self.func).nonlocals
        return self._nonlocals

    # Method for obtaining the function source
    def get_source(self):
        try:
            source = inspect.getsource(self.func)
        except Exception:
            # inspect.getsource() fails in the interactive REPL.
            # The dill package can hack around this limitation.
            try:
                import dill
            except ModuleNotFoundError:
                warnings.warn(
                    (
                        f'To use cycept.jit interactively, '
                        f'please install the dill Python package:\n'
                        f'    {sys.executable} -m pip install dill\n'
                    ),
                    RuntimeWarning,
                )
                raise
            source = dill.source.getsource(self.func)
        # Dedent source lines
        lines = source.split('\n')
        indentation = ' ' * (len(lines[0]) - len(lines[0].lstrip()))
        if indentation:
            for i, line in enumerate(lines):
                lines[i] = line.removeprefix(indentation)
            source = '\n'.join(lines)
        # Convert lambda function to def function
        if self.func.__name__ == '<lambda>':
            # Extract just the lambda expression
            # (both inspect.getsource() and dill.source.getsource()
            # returns the surrounding code as well in the case of unassigned
            # lambda expressions).
            class LambdaVisitor(ast.NodeVisitor):
                def __init__(self, root):
                    self.root = root
                    self._lambda = None
                def get_lambda(self):
                    self._lambda = None
                    self.visit(self.root)
                    return self._lambda
                def visit_Lambda(self, node):
                    if self._lambda is None:
                        self._lambda = node
            source = ast.unparse(LambdaVisitor(ast.parse(source)).get_lambda())
            # Transform to def function
            indentation = ' ' * 4
            source = re.subn(
                r'.*(^|\W)lambda (.+?): ?',
                rf'def {self.func_name}(\g<2>):\n{indentation}return ',
                source,
                1,
            )[0]
        else:
            # Do AST round-trip, ensuring canonical code style
            # (e.g. 4 space indentation).
            source = ast.unparse(ast.parse(source))
        # Remove jit decorator and decorators above it
        from .core import jit
        lines = source.split('\n')
        for i, line in enumerate(lines):
            if match := re.search(r'^@(.+?)(\(|$)', line):
                deco_name = match[1].strip()
                try:
                    deco_func = eval(deco_name, self.module_dict)
                except Exception:
                    continue
                if deco_func is jit:
                    lines = lines[i+1:]
                    break
            elif line.startswith('def '):
                # Arrived at definition line without finding jit decorator.
                # This happens if the "decoration" is done as
                # func = jit(func) rather than using the @jit syntax.
                break
        source = '\n'.join(lines)
        return source

    # Method for extracting user-defined type annotations.
    # The return value is the annotations as well as
    # the modified source without the annotations.
    def extract_annotations(self):
        class AnnotationExtractor(ast.NodeTransformer):
            def __init__(self, call):
                self.call = call
                self.annotations = {}
            def extract(self):
                self.annotations.clear()
                source = ast.unparse(self.visit(self.call.ast))
                return self.annotations, source
            def add_annotation(self, name, tp):
                if not isinstance(name, str):
                    name = ast.unparse(name)
                tp = eval(ast.unparse(tp), self.call.module_dict)
                tp_old = self.annotations.setdefault(name, tp)
                if tp_old == tp:
                    return
                warnings.warn(
                    (
                        f'Name {name!r} annotated with different types '
                        f'in jitted {self.call}: {tp}, {tp_old}. '
                        f'The first will be used'
                    ),
                    RuntimeWarning,
                )
            def visit_arg(self, node):
                if not node.annotation:
                    return node
                self.add_annotation(node.arg, node.annotation)
                return type(node)(**(vars(node) | {'annotation': None}))
            def visit_FunctionDef(self, node):
                node = self.generic_visit(node)
                if not node.returns:
                    return node
                self.add_annotation('return', node.returns)
                return type(node)(**(vars(node) | {'returns': None}))
            def visit_AnnAssign(self, node):
                if node.simple:
                    self.add_annotation(node.target, node.annotation)
                else:
                    warnings.warn(
                        (
                            f'Ignoring complex type annotation '
                            f'{ast.unparse(node)!r} in jitted {self.call}'
                        ),
                        RuntimeWarning,
                    )
                if not node.value:
                    # Annotation-only statement. Remove completely.
                    return None
                # Transform to assignment statement without annotation
                return ast.Assign(**(
                    vars(node) | {'targets': [node.target], 'value': node.value}
                ))
        annotations, source = AnnotationExtractor(self).extract()
        return annotations, source

    # Method for transforming all Unicode names to ASCII names
    def asciify_names(self):
        from .core import asciify
        class UnicodeRenamer(ast.NodeTransformer):
            def __init__(self, call):
                self.call = call
            def rename(self):
                source = ast.unparse(self.visit(self.call.ast))
                return source
            def visit_Name(self, node):
                name = asciify(node.id)
                return type(node)(**(vars(node) | {'id': name}))
            def visit_ClassDef(self, node):
                node = self.generic_visit(node)
                name = asciify(node.name)
                return type(node)(**(vars(node) | {'name': name}))
            def visit_FunctionDef(self, node):
                node = self.generic_visit(node)
                name = asciify(node.name)
                return type(node)(**(vars(node) | {'name': name}))
            def visit_arg(self, node):
                node = self.generic_visit(node)
                name = asciify(node.arg)
                return type(node)(**(vars(node) | {'arg': name}))
            def visit_keyword(self, node):
                node = self.generic_visit(node)
                name = asciify(node.arg)
                return type(node)(**(vars(node) | {'arg': name}))
            def visit_Attribute(self, node):
                node = self.generic_visit(node)
                name = asciify(node.attr)
                return type(node)(**(vars(node) | {'attr': name}))
        self._source = UnicodeRenamer(self).rename()
        # Note that name changes caused by the above ASCIIfication
        # makes the AST, the function signature and many other attributes
        # invalid, yet we do not flag any attributes as such.

    # Method for wrapping NumPy array variables in numpy.asarray()
    # whenever operations are used that do not work with Cython memoryviews.
    def protect_arrays(self, array_args):
        from .core import NdarrayTypeInfo
        # Find local NumPy arrays
        names = {
            name
            for name, tp in self.locals_types.items()
            if isinstance(tp, NdarrayTypeInfo)
        }
        names.discard('return')
        if not names:
            # No local arrays found
            return
        # Wrap local arrays
        class ArrayWrapper(ast.NodeTransformer):
            class ArrayAttribute(typing.NamedTuple):
                name: str
                is_scalar: bool = False
                is_sequence: bool = False
            array_attributes = {
                'shape': ArrayAttribute('shape', is_sequence=True),
                'strides': ArrayAttribute('strides', is_sequence=True),
                'ndim': ArrayAttribute('ndim', is_scalar=True),
                'size': ArrayAttribute('size', is_scalar=True),
                'itemsize': ArrayAttribute('itemsize', is_scalar=True),
                'nbytes': ArrayAttribute('nbytes', is_scalar=True),
            }
            def __init__(self, call, names, array_args, wrapper_func, tmp_name):
                self.call = call
                self.names = names
                self.array_args = array_args
                self.wrapper_func = wrapper_func
                self.tmp_name = tmp_name
                self.sequence_attrs = set()
                self.wrapped_any = False
                self.tmp_any = False
            def wrap(self):
                self.sequence_attrs.clear()
                self.wrapped_any = False
                self.tmp_any = False
                return self.visit(self.call.ast)
            def wrap_node(self, node, kind=0):
                nodes_subscript = []
                while isinstance(node, ast.Subscript):
                    nodes_subscript.append(node)
                    node = node.value
                if not isinstance(node, ast.Name) or node.id not in self.names:
                    return
                if nodes_subscript:
                    # Memoryview/array indexing arr[x0][x1, x2][...].
                    # If we can prove that all x are scalars
                    # we use the memoryview as is.
                    for node_subscript in nodes_subscript:
                        if isinstance(node_subscript.slice, ast.Tuple):
                            if not all(self.is_scalar(el) for el in node_subscript.slice.elts):
                                break
                        elif not self.is_scalar(node_subscript.slice):
                            break
                    else:
                        return
                # Wrap node
                self.wrapped_any = True
                if kind == 0:
                    node.id = f'{self.wrapper_func}({node.id})'
                elif kind == 1:
                    self.tmp_any = True
                    node.id = f'{self.tmp_name} = {self.wrapper_func}({node.id}); {self.tmp_name}'
            def is_scalar(self, node):
                if isinstance(node, ast.Name):
                    tp = self.call.types[node.id]
                    return tp.startswith('cython.') and '[' not in tp
                elif isinstance(node, ast.Constant):
                    return isinstance(node.value, int)
                elif isinstance(node, ast.UnaryOp):
                    return self.is_scalar(node.operand)
                elif isinstance(node, ast.BinOp):
                    return self.is_scalar(node.left) and self.is_scalar(node.right)
                elif isinstance(node, ast.Subscript):
                    if isinstance(node.value, ast.Name):
                        return node.value.id in self.names and self.is_scalar(node.slice)
                    elif isinstance(node.value, ast.Attribute):
                        return (
                            node.value.value.id in self.names
                            and (array_attribute := self.array_attributes.get(node.value.attr)) is not None
                            and array_attribute.is_sequence
                            and self.is_scalar(node.slice)
                        )
                elif isinstance(node, ast.Attribute):
                    return (
                        isinstance(node.value, ast.Name)
                        and node.value.id in self.names
                        and (array_attribute := self.array_attributes.get(node.attr)) is not None
                        and array_attribute.is_scalar
                    )
                return False
            def visit_UnaryOp(self, node):
                node = self.generic_visit(node)
                self.wrap_node(node.operand)
                return node
            def visit_BinOp(self, node):
                node = self.generic_visit(node)
                self.wrap_node(node.left)
                self.wrap_node(node.right)
                return node
            def visit_Compare(self, node):
                node = self.generic_visit(node)
                self.wrap_node(node.left)
                for comparator in node.comparators:
                    self.wrap_node(comparator)
                return node
            def visit_AugAssign(self, node):
                node = self.generic_visit(node)
                self.wrap_node(node.target, kind=1)
                return node
            # Use NumPy arrays as arguments within function calls
            # if array_args is True.
            def visit_Call(self, node):
                node = self.generic_visit(node)
                if not self.array_args:
                    return node
                for arg in node.args:
                    self.wrap_node(arg)
                return node
            # Cython memoryviews and NumPy arrays share a subset of
            # their attributes, as specified by array_attributes.
            # Below, the visit_Attribute() method will check if an attribute
            # lookup is a recognized memoryview attribute, in which case the
            # lookup will not be wrapped. Otherwise it will. However, for the
            # sequence attributes, e.g. 'shape', the value of the Cython
            # memoryview attribute and the NumPy array attribute are not
            # equivalent (they have different lengths in general). What is
            # equivalent is the result of subsequent indexing into these
            # sequences, e.g. .shape[i], with i an integer (scalar).
            # The visit_Subscript() method below will find such subscript
            # nodes and flag them, so that the visit_Attribute() method will
            # know not to wrap these. Sequence attribute lookups that are not
            # immediately indexed will not be flagged
            # and will thus be wrapped.
            def visit_Subscript(self, node):
                value = node.value
                if (
                    isinstance(value, ast.Attribute)
                    and isinstance(value.value, ast.Name)
                    and value.value.id in self.names
                    and (array_attribute := self.array_attributes.get(value.attr)) is not None
                    and array_attribute.is_sequence
                    and self.is_scalar(node.slice)
                ):
                    # Record usage of memoryview.shape[i]
                    self.sequence_attrs.add(value)
                node = self.generic_visit(node)
                return node
            def visit_Attribute(self, node):
                node = self.generic_visit(node)
                if (array_attribute := self.array_attributes.get(node.attr)) is None:
                    # Not a memoryview attribute lookup
                    self.wrap_node(node.value)
                elif array_attribute.is_sequence and node not in self.sequence_attrs:
                    # Memoryview attribute lookup of sequence
                    # that is not immediately indexed by a scalar.
                    self.wrap_node(node.value)
                return node
        wrapper_func = '_cycept_asarray_'
        tmp_name = '_cycept_tmp_'
        array_wrapper = ArrayWrapper(self, names, array_args, wrapper_func, tmp_name)
        self._source = ast.unparse(array_wrapper.wrap())
        self._ast = None  # invalidate AST (cannot use the above as expressions are used as names)
        if not array_wrapper.wrapped_any:
            return
        if array_wrapper.tmp_any:
            from .core import record_types
            record_types(self, {tmp_name: None})
        # Add wrapper_func as local variable
        class NodeAdder(ast.NodeTransformer):
            def __init__(self, call, node_to_add):
                self.call = call
                self.node_to_add = node_to_add
            def add_node(self):
                return self.visit(self.call.ast)
            def visit_FunctionDef(self, node):
                if self.node_to_add is None:
                    return node
                node = ast.FunctionDef(**(vars(node) | {'body': [self.node_to_add] + node.body}))
                self.node_to_add = None
                return node
        self._source = ast.unparse(NodeAdder(self, ast.parse(f'{wrapper_func} = _{wrapper_func}_')).add_node())
        self._ast = None  # invalidate AST (above node addition is not strictly legal)
        # We need to assign _{wrapper_func}_ to numpy.asarray() in the extension module
        self.cython_module_lines += [
            '\n# For converting memoryviews back to NumPy arrays',
            f'from numpy import asarray as _{wrapper_func}_'
        ]

    # Method for updating the source from being a Python function
    # to being a Cython extension module.
    def to_cython(self, directives):
        from .core import asciify
        preamble = [
            '#',  # first comment line can have C code attached to it
            f'# Cycept version of {self} with type signature',
            f'# {self!r}',
        ]
        preamble += self.cython_module_lines
        excludes = (self.func_name, 'cython')
        fake_globals = [
            f'{asciify(name)} = object()'
            for name in collections.ChainMap(self.nonlocals, self.globals)
            if name not in excludes
        ]
        if fake_globals:
            preamble += [
                '\n# Declare fake globals',
                *fake_globals,
            ]
        preamble += [
            '\n# Import Cython',
            'import cython',
            'cimport cython',
        ]
        header_top = ['\n# Function to be jitted']
        if self.ccall_allowed():
            header_top.append('@cython.ccall')
        for directive, val in directives.items():
            header_top.append(f'@cython.{directive}({val!r})')
        declaration_locals = ', '.join(
            f'{name}={tp}'
            for name, tp in self.types.items()
            if name not in self.locals_excludes
        )
        declaration_return = self.types.get('return', 'object')
        header_bot = []
        if declaration_locals:
            header_bot.append(f'@cython.locals({declaration_locals})')
        header_bot.append(f'@cython.returns({declaration_return})')
        # ASCIIfy Unicode variable names in the source and bottom header
        self._source = '\n'.join(
            itertools.chain(
                header_bot,
                self.source.split('\n'),
            )
        )
        self._ast = None
        self.asciify_names()
        # Combine with preamble and top header
        self._source = '\n'.join(
            itertools.chain(
                preamble,
                header_top,
                self.source.split('\n'),
            )
        )
        # ASCIIfy the stored function name itself
        self._func_name = asciify(self.func_name)

    # Method for determining whether the function to be Cythonized
    # may be done so using @cython.ccall (cpdef).
    def ccall_allowed(self):
        # *args and **kwargs not allowed with ccall
        for param in self.signature.parameters.values():
            if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
                return False
        # Closures not allowed with ccall
        class ClosureVisitor(ast.NodeVisitor):
            def __init__(self, call):
                self.call = call
                self._contains_closure = False
            def contains_closure(self):
                self._contains_closure = False
                self.visit(self.call.ast)
                return self._contains_closure
            def visit_FunctionDef(self, node):
                self.generic_visit(node)
                if node.name != self.call.func_name:
                    self._contains_closure = True
            def visit_Lambda(self, node):
                self._contains_closure = True
        if ClosureVisitor(self).contains_closure():
            return False
        # No violations found
        return True

    # Method for handling Cythonization and C compilation
    def compile(self, optimizations, c_lang, html, silent):
        from .core import asciify
        # Cythonize and compile extension module within temporary directory.
        # The compiled module is then imported (through hacking of sys.path)
        # before it is removed from disk.
        @contextlib.contextmanager
        def hack_sys_path():
            sys.path.append(dir_name)
            try:
                yield
            except Exception:
                raise
            finally:
                sys.path.remove(dir_name)
        module_name = f'_cycept_module_{self.hash}'
        namespace = {}
        source_c = None
        html_annotation = None
        with (
            tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as dir_name,
            hack_sys_path(),
        ):
            # Write Cython source to file
            module_path = pathlib.Path(dir_name) / module_name
            module_path.with_suffix('.pyx').write_text(self.source, 'utf-8')
            # Call Cython. We do so within a subprocess in order
            # to capture stdout an stderr when running silently.
            cmd = [
                sys.executable,
                '-c',
                '; '.join([
                    'import cycept.compile',
                    'cycept.compile.compile({})'
                    .format(', '.join([
                        f'{str(module_path)!r}',
                        f'{optimizations!r}',
                        f'{c_lang!r}',
                        f'{html!r}',
                    ]))
                ]),
            ]
            run_kwargs = {}
            if silent:
                run_kwargs |= {
                    'stdout': subprocess.PIPE,
                    'stderr': subprocess.STDOUT,
                    'text': True,
                }
            tic = time.perf_counter()
            cproc = subprocess.run(cmd, **run_kwargs)
            toc = time.perf_counter()
            if cproc.returncode != 0:
                if sys.flags.interactive:
                    # Do not remove the compilation files immediately when
                    # running interactively, allowing the user to inspect them.
                    input(f'Press Enter to clean up temporary build directory {dir_name} ')
                msg = ['Cythonization failed.']
                if silent:
                    msg += ['Subprocess output:', cproc.stdout]
                raise OSError('\n'.join(msg))
            # Import function from compiled module into temporary namespace
            exec(f'import {module_name}', namespace)
            # Read in C source and annotated HTML
            for ext in ['c', 'cc', 'cpp', 'cxx']:
                if (path_c := module_path.with_suffix(f'.{ext}')).is_file():
                    source_c = path_c.read_text('utf-8')
                    break
            else:
                ext = None
            if (path_html := module_path.with_suffix('.html')).is_file():
                html_annotation = path_html.read_text('utf-8')
        # Extract compiled function
        module_compiled = namespace[module_name]
        module_compiled_dict = module_compiled.__dict__
        func_compiled = module_compiled_dict[self.func_name]
        # Replace fake globals and non-locals with actual
        # globals and non-locals within extension module.
        for name, val in collections.ChainMap(self.nonlocals, self.globals).items():
            if name == self.func_name:
                continue
            module_compiled_dict[name] = val
            module_compiled_dict[asciify(name)] = val
        # Store compilation products
        self.compiled = self.Compiled(
            func_compiled,
            module_compiled,
            source_c,
            ext,
            html_annotation,
        )
        # Return compilation time (in seconds)
        return toc - tic

    # Method for viewing the annotated HTML
    def __call__(self, dir_name=None):
        if self.compiled is None or self.compiled.html is None:
            print(f'No Cython HTML annotation generated for {self.func_name}()')
            return
        if dir_name is None:
            dir_name = tempfile.mkdtemp()  # not cleaned up
        module_path = pathlib.Path(dir_name) / self.compiled.module.__name__
        if self.compiled.source is not None and self.compiled.source_ext is not None:
            module_path.with_suffix(f'.{self.compiled.source_ext}').write_text(
                self.compiled.source,
                'utf-8',
             )
        path_html = module_path.with_suffix('.html')
        path_html.write_text(self.compiled.html, 'utf-8')
        webbrowser.open_new_tab(str(path_html.as_uri()))
        return dir_name

    # Method for pretty printing (showing types)
    def __repr__(self):
        from .core import prettify_type
        return f'{self.func_name}({{}})'.format(
            ', '.join(
                f'{name}: {prettify_type(tp)}'
                for name, tp in zip(self.arguments, self.arguments_types)
            )
        )

    # Method for pretty printing (showing module)
    def __str__(self):
        return f'function {self.func_name}() defined in {self.module}'




import contextlib
import sys

from .test_bench import bench


def test(kind=None):
    testfiles = ['cycept.tests']
    if isinstance(kind, str):
        kind = [kind]
    if isinstance(kind, (list, tuple)):
        kind = [k.removeprefix('test_') for k in kind]
        testfiles = [f'cycept.tests.test_{k}' for k in kind]
    # Run pytest on the cycept.test submodule.
    # We prevent pytest from writing to __pycache__ by hacking on
    # sys.dont_write_bytecode. We do this as such files are not cleanly
    # removed by 'pip uninstall cycept'.
    @contextlib.contextmanager
    def hack_sys_dont_write_bytecode():
        backup = sys.dont_write_bytecode
        sys.dont_write_bytecode = True
        try:
            yield
        except Exception:
            raise
        finally:
            sys.dont_write_bytecode = backup
    with hack_sys_dont_write_bytecode():
        import pytest
        pytest.main(['-p', 'no:cacheprovider', '--pyargs', *testfiles])




from .core import jit, NdarrayTypeInfo, __version__
from .tests import bench, test


# Function for checking whether Cycept functions correctly
def check(silent_if_ok=False):
    import sys
    import timeit
    import numpy as np
    def get_test_func():
        def wallis(n):
            π = 2
            for i in range(1, n):
                π *= 4 * i ** 2 / (4 * i ** 2 - 1)
            return π
        return wallis
    def timing(func, n, calls=100):
        return min(timeit.repeat(lambda: func(n), number=calls))
    func = get_test_func()
    n = 10_000
    ok = True
    try:
        func_jitted = jit(func, silent=True)
        func_jitted(n)
    except Exception:
        ok = False
    if not ok:
        print('Cycept failed to compile test function', file=sys.stderr)
    if ok and not np.isclose(func(n), func_jitted(n)):
        ok = False
        print(
            'Cycept compiled the test function without errors,',
            'but the return value of jitted function is inconsistent',
            'with that of the pure Python version',
            file=sys.stderr,
        )
    if ok:
        if timing(func_jitted, n) > timing(func, n):
            ok = False
            print(
                'Cycept compiled the test function without errors,',
                'but the jitted function is slower than the pure Python version',
                file=sys.stderr,
            )
    if ok:
        if not silent_if_ok:
            print('Cycept functions correctly')
    else:
        print('Below we attempt compilation once more', file=sys.stderr)
        func = get_test_func()
        jit(func, silent=False)(n)
        raise RuntimeError('Cycept failure')

# Add functionality (or at least allow their syntax to be used)
# to the Python standard library that is missing
# due to the Python version being too low.
def _implement_future():
    import contextlib
    import inspect
    import os
    import tempfile

    # Implement contextlib.chdir
    # (available from Python 3.11).
    if not hasattr(contextlib, 'chdir'):
        @contextlib.contextmanager
        def chdir(path):
            cwd_backup = os.getcwd()
            os.chdir(path)
            try:
                yield
            except Exception:
                raise
            finally:
                os.chdir(cwd_backup)
        contextlib.chdir = chdir

    # Allow tempfile.TemporaryDirectory to take
    # ignore_cleanup_errors as an argument
    # (available from Python 3.10).
    if 'ignore_cleanup_errors' not in inspect.signature(
        tempfile.TemporaryDirectory
    ).parameters:
        class TemporaryDirectory(tempfile.TemporaryDirectory):
            def __init__(self, *args, ignore_cleanup_errors=False, **kwargs):
                super().__init__(*args, *kwargs)
        tempfile.TemporaryDirectory = TemporaryDirectory

_implement_future()

